# -*- coding: utf-8 -*-
# https://github.com/python-poetry/poetry/pull/2366#issuecomment-652418094
from nanocompore.common import get_version

__version__ = get_version()

__description__ = 'Software package that identifies raw signal changes between two conditions from https://github.com/jts/nanopolish resquiggled dRNA-Seq data.'
